static inline void
mat_ident(float *mat)
{
   static const float ident[] =
     {
        1, 0, 0, 0,
        0, 1, 0, 0,
        0, 0, 1, 0,
        0, 0, 0, 1
     };
   memcpy(mat, ident, sizeof(ident));
}

static inline void
mat_mul(float *dst, float *mat1, float *mat2)
{
   float res[16];
   int i, j;

   for (i = 0; i < 4; i++)
     {
        for (j = 0; j < 4; j++)
          {
             res[(4 * i) + j] =
               (mat1[     j] * mat2[(4 * i)    ]) +
               (mat1[ 4 + j] * mat2[(4 * i) + 1]) +
               (mat1[ 8 + j] * mat2[(4 * i) + 2]) +
               (mat1[12 + j] * mat2[(4 * i) + 3]);
          }
     }
   memcpy(dst, res, sizeof(res));
}

static inline void
mat_trans(float *mat, float x, float y, float z)
{
   float tmp[16];
   mat_ident(tmp);
   tmp[12] = x;
   tmp[13] = y;
   tmp[14] = z;
   mat_mul(mat, tmp, mat);
}

static inline void
mat_scale(float *mat, float x, float y, float z)
{
   float tmp[16];
   mat_ident(tmp);
   tmp[ 0] = x;
   tmp[ 5] = y;
   tmp[10] = z;
   mat_mul(mat, tmp, mat);
}

static inline float
mat_deg_rad(float deg)
{
   return (deg * M_PI * 2.0) / 360.0;
}

static inline void
mat_rot_x(float *mat, float a)
{
   float tmp[16];
   mat_ident(tmp);
   tmp[ 5] =  cos(mat_deg_rad(a));
   tmp[ 9] = -sin(mat_deg_rad(a));
   tmp[ 6] =  sin(mat_deg_rad(a));
   tmp[10] =  cos(mat_deg_rad(a));
   mat_mul(mat, tmp, mat);
}

static inline void
mat_rot_y(float *mat, float a)
{
   float tmp[16];
   mat_ident(tmp);
   tmp[ 0] =  cos(mat_deg_rad(a));
   tmp[ 8] =  sin(mat_deg_rad(a));
   tmp[ 2] = -sin(mat_deg_rad(a));
   tmp[10] =  cos(mat_deg_rad(a));
   mat_mul(mat, tmp, mat);
}

static inline void
mat_rot_z(float *mat, float a)
{
   float tmp[16];
   mat_ident(tmp);
   tmp[ 0] =  cos(mat_deg_rad(a));
   tmp[ 4] = -sin(mat_deg_rad(a));
   tmp[ 1] =  sin(mat_deg_rad(a));
   tmp[ 5] =  cos(mat_deg_rad(a));
   mat_mul(mat, tmp, mat);
}

static inline void
mat_frustum(float *mat, float l, float r, float b, float t, float z0, float zf)
{
   float tmp, xdist, ydist, zdist;

   tmp = 2.0 * z0;
   xdist = r - l;
   ydist = t - b;
   zdist = zf - z0;
   mat_ident(mat);
   mat[ 0] = tmp / xdist;
   mat[ 5] = tmp / ydist;
   mat[ 8] = (r + l) / xdist;
   mat[ 9] = (t + b) / ydist;
   mat[10] = (-zf - z0) / zdist;
   mat[11] = -1;
   mat[14] = (-tmp * zf) / zdist;
   mat[15] = 0;
}

static inline void
mat_persp(float *mat, float fov, float aspect, float z0, float zf)
{
   float ymax, xmax;

   ymax = z0 * tanf((fov * M_PI) / 360.0);
   xmax = ymax * aspect;
   mat_frustum(mat, -xmax, xmax, -ymax, ymax, z0, zf);
}

static inline void
mat_ortho(float *mat, int vw, int vh, int vflip)
{
   float l = 0.0, r = 0.0, t = 0.0, b = 0.0;
   float cosv = 1.0, sinv = 0.0;
   float near = -1000.0, far = 1000.0;

   r = vw;
   if (!vflip) b = vh;
   else t = vh;

   mat[ 0] = (2.0 / (r - l)) * ( cosv);
   mat[ 1] = (2.0 / (r - l)) * ( sinv);
   mat[ 2] = 0.0;
   mat[ 3] = 0.0;

   mat[ 4] = (2.0 / (t - b)) * (-sinv);
   mat[ 5] = (2.0 / (t - b)) * ( cosv);
   mat[ 6] = 0.0;
   mat[ 7] = 0.0;

   mat[ 8] = 0.0;
   mat[ 9] = 0.0;
   mat[10] = -(2.0 / (far - near));
   mat[11] = 1.0;

   mat[12] = -((r + l) / (r - l));
   mat[13] = -((t + b) / (t - b));
   mat[14] = -((near + far) / (far - near));
   mat[15] = 1.0;
}
